︠c7f857f6-7273-4145-98b7-e24cc9019345i︠
%html
<p>This worksheet implements the Euler-Maruyama method for SDEs (see <a href="http://en.wikipedia.org/wiki/Euler-Maruyama_method">http://en.wikipedia.org/wiki/Euler-Maruyama_method</a> for more details) and presents an interactive demo of a simple dynamical system exhibiting a Hopf bifurcation with noise (taken from <a href="http://johncarlosbaez.wordpress.com/2010/12/24/this-weeks-finds-week-308/">http://johncarlosbaez.wordpress.com/2010/12/24/this-weeks-finds-week-308/</a>).</p>
<p>Other uses of the Euler-Maruyama method can be found in the Sage <a href="http://wiki.sagemath.org/interact/diffeq#Euler-MaruyamamethodandgeometricBrownianmotion.28acommonsimplemodelofthestockmarket.29">interact/diffeq demo</a>.&nbsp; In fact, this worksheet was heavily inspired by that demo.</p>

︡da13bf6a-a7ec-487a-8c19-6fc6002dde5c︡{"html": "<p>This worksheet implements the Euler-Maruyama method for SDEs (see <a href=\"http://en.wikipedia.org/wiki/Euler-Maruyama_method\">http://en.wikipedia.org/wiki/Euler-Maruyama_method</a> for more details) and presents an interactive demo of a simple dynamical system exhibiting a Hopf bifurcation with noise (taken from <a href=\"http://johncarlosbaez.wordpress.com/2010/12/24/this-weeks-finds-week-308/\">http://johncarlosbaez.wordpress.com/2010/12/24/this-weeks-finds-week-308/</a>).</p>\n<p>Other uses of the Euler-Maruyama method can be found in the Sage <a href=\"http://wiki.sagemath.org/interact/diffeq#Euler-MaruyamamethodandgeometricBrownianmotion.28acommonsimplemodelofthestockmarket.29\">interact/diffeq demo</a>.&nbsp; In fact, this worksheet was heavily inspired by that demo.</p>"}︡
︠b0066842-1975-495c-9b06-b07e101a770e︠
def euler_maruyama(a, b, x0, h, N):
    """
    Implements the Euler-Maruyama method for SDEs in 2D of the form 
    dX = a(X)dt + b(X)dW, where W is a vector of independent 
    Brownian motions.
    
    INPUT:
        
        - ``a``, ``b`` - vector-valued functions 
        - ``x0`` - initial conditions (2D vector)
        - ``h`` - step size of the method
        - ``N``  - number of steps to take
        
    OUTPUT: List of integration points.
    """
    
    output = [x0]
    for _ in range(N):
        W = vector([normalvariate(0, sqrt(h)), normalvariate(0, sqrt(h))])
        output.append(output[-1] + h*a(output[-1]) + b(output[-1]).pairwise_product(W))
    return output
︡68902dd8-390b-4108-9afc-0d12dfb8e90d︡︡
︠02588fc4-2f96-4aee-9c08-43820baaa96bi︠
%html
<p>The interactive demo allows you to control the parameter $\beta$ (determining the limit cycle) and the amplitude of the noise $\ell$.&nbsp; When the noise is absent ($\ell = 0$), the origin is a stable fixed point when $\beta &lt; 0$.&nbsp; As $\beta$ becomes positive, the origin turns into an unstable fixed point and a stable limit cycle is born at $r = \sqrt{\beta}$.&nbsp;</p>
<p>When noise is added to the system, the limit cycle can still be discerned for $\beta &gt; 0$.&nbsp;</p>
<p>For those of us schooled in classical ODEs, it is worth keeping in mind that these trajectories are continuous but not differentiable.&nbsp; The EM solver computes only a finite number of points on each trajectory and then plots a straight line between the individual points, giving the appearance of a curve which is "somewhat smooth" for large step sizes.&nbsp; The step size can of course be lowered, revealing the intricate features of the trajectories at the expense of longer computations.</p>

︡2479d784-fa49-4bcb-a70b-d4182c1027ff︡{"html": "<p>The interactive demo allows you to control the parameter $\\beta$ (determining the limit cycle) and the amplitude of the noise $\\ell$.&nbsp; When the noise is absent ($\\ell = 0$), the origin is a stable fixed point when $\\beta &lt; 0$.&nbsp; As $\\beta$ becomes positive, the origin turns into an unstable fixed point and a stable limit cycle is born at $r = \\sqrt{\\beta}$.&nbsp;</p>\n<p>When noise is added to the system, the limit cycle can still be discerned for $\\beta &gt; 0$.&nbsp;</p>\n<p>For those of us schooled in classical ODEs, it is worth keeping in mind that these trajectories are continuous but not differentiable.&nbsp; The EM solver computes only a finite number of points on each trajectory and then plots a straight line between the individual points, giving the appearance of a curve which is \"somewhat smooth\" for large step sizes.&nbsp; The step size can of course be lowered, revealing the intricate features of the trajectories at the expense of longer computations.</p>"}︡
︠c1948fb3-7e20-4071-82e5-a3ff3fb5dd15︠
@interact
def hopf_sde(beta=slider(srange(-2, 2, .1), default=1.0), ell=slider(srange(0, 1, .1), default=0.1)):      
    h = .02 # Step size      
    hopf_a = lambda x : vector([-x[1], x[0]]) + beta*x - x.norm()^2*x
    hopf_b = lambda x : vector([ell, ell])
    output1 = euler_maruyama(hopf_a, hopf_b, vector([0.0, 0.1]), h, int(10/h))
    output2 = euler_maruyama(hopf_a, hopf_b, vector([0.0, 2.0]), h, int(10/h))
    p =  list_plot([p.list() for p in output1], plotjoined=True, thickness=1, color='red')
    p += list_plot([p.list() for p in output2], plotjoined=True, thickness=1, color='green')
    p.show(xmin=-2, xmax=2, ymin=-2, ymax=2)
︡c9ff4bd7-73cb-451e-bad2-1425468fa034︡{"html": "<!--notruncate--><div padding=6 id=\"div-interact-10\"> <table width=800px height=20px bgcolor=\"#c5c5c5\"\n                 cellpadding=15><tr><td bgcolor=\"#f9f9f9\" valign=top align=left><table>\n<tr><td colspan=3><table><tr><td align=right><font color=\"black\">beta&nbsp;</font></td><td><table><tr><td>\n        <div id=\"slider-beta-10\" style=\"margin:0px; margin-left: 1.0em; margin-right: 1.0em; width: 15.0em;\"></div>\n        </td><td><font color=\"black\" id=\"slider-beta-10-lbl\"></font></td></tr></table><script>(function(){ var values = [\"-2.00000000000000\",\"-1.90000000000000\",\"-1.80000000000000\",\"-1.70000000000000\",\"-1.60000000000000\",\"-1.50000000000000\",\"-1.40000000000000\",\"-1.30000000000000\",\"-1.20000000000000\",\"-1.10000000000000\",\"-0.999999999999999\",\"-0.899999999999999\",\"-0.799999999999999\",\"-0.699999999999999\",\"-0.599999999999999\",\"-0.499999999999999\",\"-0.399999999999999\",\"-0.299999999999999\",\"-0.199999999999999\",\"-0.0999999999999994\",\"6.38378239159465e-16\",\"0.100000000000001\",\"0.200000000000001\",\"0.300000000000001\",\"0.400000000000001\",\"0.500000000000001\",\"0.600000000000001\",\"0.700000000000001\",\"0.800000000000001\",\"0.900000000000001\",\"1.00000000000000\",\"1.10000000000000\",\"1.20000000000000\",\"1.30000000000000\",\"1.40000000000000\",\"1.50000000000000\",\"1.60000000000000\",\"1.70000000000000\",\"1.80000000000000\",\"1.90000000000000\"]; setTimeout(function() {\n    $('#slider-beta-10').slider({\n        step: 1, min: 0, max: 39, value: 30,\n        change: function (e,ui) { var position = ui.value; if(values!=null) $('#slider-beta-10-lbl').text(values[position]); interact(10, '_interact_.update(\\'10\\', \\'beta\\', 1, _interact_.standard_b64decode(\\''+encode64(position)+'\\'), globals()); _interact_.recompute(\\'10\\');'); },\n        slide: function(e,ui) { if(values!=null) $('#slider-beta-10-lbl').text(values[ui.value]); }\n    });\n    if(values != null) $('#slider-beta-10-lbl').text(values[$('#slider-beta-10').slider('value')]);\n    }, 1); })();</script></td>\n</tr><tr><td align=right><font color=\"black\">ell&nbsp;</font></td><td><table><tr><td>\n        <div id=\"slider-ell-10\" style=\"margin:0px; margin-left: 1.0em; margin-right: 1.0em; width: 15.0em;\"></div>\n        </td><td><font color=\"black\" id=\"slider-ell-10-lbl\"></font></td></tr></table><script>(function(){ var values = [\"0.000000000000000\",\"0.100000000000000\",\"0.200000000000000\",\"0.300000000000000\",\"0.400000000000000\",\"0.500000000000000\",\"0.600000000000000\",\"0.700000000000000\",\"0.800000000000000\",\"0.900000000000000\"]; setTimeout(function() {\n    $('#slider-ell-10').slider({\n        step: 1, min: 0, max: 9, value: 1,\n        change: function (e,ui) { var position = ui.value; if(values!=null) $('#slider-ell-10-lbl').text(values[position]); interact(10, '_interact_.update(\\'10\\', \\'ell\\', 2, _interact_.standard_b64decode(\\''+encode64(position)+'\\'), globals()); _interact_.recompute(\\'10\\');'); },\n        slide: function(e,ui) { if(values!=null) $('#slider-ell-10-lbl').text(values[ui.value]); }\n    });\n    if(values != null) $('#slider-ell-10-lbl').text(values[$('#slider-ell-10').slider('value')]);\n    }, 1); })();</script></td>\n</tr></table></td></tr>\n<tr><td></td><td style='width: 100%;'><div id=\"cell-interact-10\"><?__SAGE__START>\n        <table border=0 bgcolor=\"white\" width=100%>\n        <tr><td bgcolor=\"white\" align=left valign=top><pre><?__SAGE__TEXT></pre></td></tr>\n        <tr><td  align=left valign=top><?__SAGE__HTML></td></tr>\n        </table><?__SAGE__END></div></td><td></td></tr>\n<tr><td colspan=3></td></tr>\n</table></td>\n                 </tr></table></div>\n                 "}︡
